# Modified from Isaac Lab Project Developers (https://github.com/isaac-sim/IsaacLab/blob/main/CONTRIBUTORS.md).
# SPDX-License-Identifier: BSD-3-Clause

import os

import numpy as np


class MotionSequence:
    """
    A sequence of motion data.

    Modified from https://github.com/isaac-sim/IsaacLab/blob/main/source/isaaclab_tasks/isaaclab_tasks/direct/humanoid_amp/motions/motion_loader.py.

    This class is modified to use numpy instead of torch to load the motion data, suitable for CPU-only environment.
    """

    @classmethod
    def load(cls, path: str) -> "MotionSequence":
        """Load a motion file and return a MotionSequence object.

        Args:
            path: Motion file path to load.

        Raises:
            AssertionError: If the specified motion file doesn't exist.
        """
        assert os.path.isfile(path), f"Invalid file path: {path}"
        data = np.load(path)

        num_frames = np.max([data["dof_positions"].shape[0], data["body_positions"].shape[0]])

        motion = cls(
            num_frames=num_frames,
            dof_names=data["dof_names"].tolist(),
            body_names=data["body_names"].tolist(),
            fps=data["fps"],
        )

        motion._dof_positions[:] = data["dof_positions"]     # joint positions, rad
        motion._dof_velocities[:] = data["dof_velocities"]   # joint velocities, rad/s
        motion._body_positions[:] = data["body_positions"]   # link positions, m
        motion._body_rotations[:] = data["body_rotations"]   # link rotations, (qw, qx, qy, qz) quaternion
        motion._body_linear_velocities[:] = data["body_linear_velocities"]       # link linear velocities, m/s
        motion._body_angular_velocities[:] = data["body_angular_velocities"]     # link angular velocities, rad/s

        print(f"Motion loaded ({path}): duration: {motion._duration} sec, # of frames: {motion._num_frames}, FPS: {motion._fps}")

        return motion

    def __init__(self, num_frames: int, dof_names: list[str], body_names: list[str], fps: int = 50) -> None:
        """Initialize a MotionSequence object.
        
        Args:
            num_frames: Number of frames.
            dof_names: List of joint names.
            body_names: List of rigid body names.
            fps: the FPS of the motion data.
        """
        self._fps = fps                                         # frame rate
        self._dt = 1.0 / self._fps                              # per frame time step, s
        self._num_frames = num_frames                           # total number of frames
        self._duration = self._dt * (self._num_frames - 1)      # duration, s
        self._dof_names = dof_names                             # joint names
        self._body_names = body_names                           # rigid body names

        # joint positions, rad
        self._dof_positions = np.zeros((self._num_frames, len(self._dof_names)), dtype=np.float32)
        # joint velocities, rad/s
        self._dof_velocities = np.zeros((self._num_frames, len(self._dof_names)), dtype=np.float32)
        # link positions, m
        self._body_positions = np.zeros((self._num_frames, len(self._body_names), 3), dtype=np.float32)
        # link rotations, (qw, qx, qy, qz) quaternion
        self._body_rotations = np.zeros((self._num_frames, len(self._body_names), 4), dtype=np.float32)
        # link linear velocities, m/s
        self._body_linear_velocities = np.zeros((self._num_frames, len(self._body_names), 3), dtype=np.float32)
        # link angular velocities, rad/s
        self._body_angular_velocities = np.zeros((self._num_frames, len(self._body_names), 3), dtype=np.float32)

        # ensure correct default quaternion representation
        self._body_rotations[:, :, 0] = 1.0

    @property
    def fps(self) -> int:
        """Frames per second."""
        return self._fps

    @property
    def dt(self) -> float:
        """Time step."""
        return self._dt

    @property
    def num_frames(self) -> int:
        """Number of frames."""
        return self._num_frames

    @property
    def duration(self) -> float:
        """Duration."""
        return self._duration

    @property
    def dof_names(self) -> list[str]:
        """Joint names."""
        return self._dof_names

    @property
    def body_names(self) -> list[str]:
        """Rigid body names."""
        return self._body_names

    @property
    def num_dofs(self) -> int:
        """Number of joints."""
        return len(self._dof_names)

    @property
    def num_bodies(self) -> int:
        """Number of rigid bodies."""
        return len(self._body_names)

    @property
    def dof_positions(self) -> np.ndarray:
        """Joint positions."""
        return self._dof_positions

    @property
    def dof_velocities(self) -> np.ndarray:
        """Joint velocities."""
        return self._dof_velocities

    @property
    def body_positions(self) -> np.ndarray:
        """Rigid body positions."""
        return self._body_positions

    @property
    def body_rotations(self) -> np.ndarray:
        """Rigid body rotations."""
        return self._body_rotations

    @property
    def body_linear_velocities(self) -> np.ndarray:
        """Rigid body linear velocities."""
        return self._body_linear_velocities

    @property
    def body_angular_velocities(self) -> np.ndarray:
        """Rigid body angular velocities."""
        return self._body_angular_velocities

    def get_dof_index(self, dof_names: list[str]) -> list[int]:
        """Get joint indexes by joint names.

        Args:
            dof_names: List of joint names.

        Raises:
            AssertionError: If the specified joint name doesn't exist.

        Returns:
            List of joint indexes.
        """
        indexes = []
        for name in dof_names:
            assert name in self._dof_names, f"The specified joint name ({name}) doesn't exist in {self._dof_names}"
            indexes.append(self._dof_names.index(name))
        return indexes

    def get_body_index(self, body_names: list[str]) -> list[int]:
        """Get rigid body indexes by rigid body names.

        Args:
            body_names: List of rigid body names.

        Raises:
            AssertionError: If the specified rigid body name doesn't exist.

        Returns:
            List of rigid body indexes.
        """
        indexes = []
        for name in body_names:
            assert name in self._body_names, f"The specified body name ({name}) doesn't exist in {self._body_names}"
            indexes.append(self._body_names.index(name))
        return indexes

    def get_frames(
        self,
        frames: int | list[int] | np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Get motion data by frame indices.
        This function will automatically clip the frame indices to the valid range. If the frame indices are out of range,
        the function will return the data of last frame (as upper bound) or first frame (as lower bound).

        Args:
            frames: List of frame indices to sample.

        Returns:
            Motion joint positions (with shape (N, num_dofs)), joint velocities (with shape (N, num_dofs)),
            rigid body positions (with shape (N, num_bodies, 3)), rigid body rotations (with shape (N, num_bodies, 4), as wxyz quaternion),
            rigid body linear velocities (with shape (N, num_bodies, 3)) and rigid body angular velocities (with shape (N, num_bodies, 3)).
        """
        frames = np.clip(frames, 0, self._num_frames - 1)

        return (
            self._dof_positions[frames],
            self._dof_velocities[frames],
            self._body_positions[frames],
            self._body_rotations[frames],
            self._body_linear_velocities[frames],
            self._body_angular_velocities[frames],
        )

    def save(self, path: str) -> None:
        """Save the motion data to a file.
        
        Args:
            path: The path to the output file.
        """
        fps = np.array([self._fps], dtype=np.int64)  # this needs to be int64
        dof_names = np.array(self._dof_names)
        body_names = np.array(self._body_names)

        np.savez(path,
            fps=fps,
            dof_names=dof_names,
            body_names=body_names,
            dof_positions=self._dof_positions,
            dof_velocities=self._dof_velocities,
            body_positions=self._body_positions,
            body_rotations=self._body_rotations,
            body_linear_velocities=self._body_linear_velocities,
            body_angular_velocities=self._body_angular_velocities,
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--file", type=str, required=True, help="Motion file")
    args, _ = parser.parse_known_args()

    motion = MotionSequence(args.file)

    print("- number of frames:", motion.num_frames)
    print("- number of DOFs:", motion.num_dofs)
    print("- number of bodies:", motion.num_bodies)



# Modified from Isaac Lab Project Developers (https://github.com/isaac-sim/IsaacLab/blob/main/CONTRIBUTORS.md).
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np


def quat_from_euler_xyz(roll: np.ndarray, pitch: np.ndarray, yaw: np.ndarray) -> np.ndarray:
    """Convert rotations given as Euler angles in radians to Quaternions.

    Note:
        The euler angles are assumed in XYZ convention (1-2-3 sequence).
        See https://ntrs.nasa.gov/citations/19770024290 for different conventions.

    Args:
        roll: Rotation around x-axis (in radians). Shape is (N,).
        pitch: Rotation around y-axis (in radians). Shape is (N,).
        yaw: Rotation around z-axis (in radians). Shape is (N,).

    Returns:
        The quaternion in (w, x, y, z). Shape is (N, 4).
    """
    sr = np.sin(roll * 0.5)
    cr = np.cos(roll * 0.5)
    sp = np.sin(pitch * 0.5)
    cp = np.cos(pitch * 0.5)
    sy = np.sin(yaw * 0.5)
    cy = np.cos(yaw * 0.5)
    # compute quaternion
    qw = -sr * sp * sy + cr * cp * cy
    qx =  sr * cp * cy + cr * sp * sy
    qy = -sr * cp * sy + cr * sp * cy
    qz =  sr * sp * cy + cr * cp * sy

    return np.stack([qw, qx, qy, qz], axis=-1)


def quat_from_euler_zyx(roll: np.ndarray, pitch: np.ndarray, yaw: np.ndarray) -> np.ndarray:
    """Convert rotations given as Euler angles in radians to Quaternions.

    Note:
        The euler angles are assumed in ZYX convention (3-2-1 sequence).
        See https://ntrs.nasa.gov/citations/19770024290 for different conventions.

    Args:
        roll: Rotation around x-axis (in radians). Shape is (N,).
        pitch: Rotation around y-axis (in radians). Shape is (N,).
        yaw: Rotation around z-axis (in radians). Shape is (N,).

    Returns:
        The quaternion in (w, x, y, z). Shape is (N, 4).
    """
    sr = np.sin(roll * 0.5)
    cr = np.cos(roll * 0.5)
    sp = np.sin(pitch * 0.5)
    cp = np.cos(pitch * 0.5)
    sy = np.sin(yaw * 0.5)
    cy = np.cos(yaw * 0.5)
    # compute quaternion
    qw = cr * cp * cy + sr * sp * sy
    qx = sr * cp * cy - cr * sp * sy
    qy = cr * sp * cy + sr * cp * sy
    qz = cr * cp * sy - sr * sp * cy

    return np.stack([qw, qx, qy, qz], axis=-1)


def quat_mul(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    """Multiply two quaternions together.

    Args:
        q1: The first quaternion in (w, x, y, z). Shape is (..., 4).
        q2: The second quaternion in (w, x, y, z). Shape is (..., 4).

    Returns:
        The product of the two quaternions in (w, x, y, z). Shape is (..., 4).

    Note:
        Supports NumPy broadcasting. If shapes don't match, broadcasting will be applied.
    """
    # Apply broadcasting to match shapes
    q1_broadcasted, q2_broadcasted = np.broadcast_arrays(q1, q2)
    
    # reshape to (N, 4) for multiplication
    shape = q1_broadcasted.shape
    q1 = q1_broadcasted.reshape(-1, 4)
    q2 = q2_broadcasted.reshape(-1, 4)
    # extract components from quaternions
    w1, x1, y1, z1 = q1[:, 0], q1[:, 1], q1[:, 2], q1[:, 3]
    w2, x2, y2, z2 = q2[:, 0], q2[:, 1], q2[:, 2], q2[:, 3]
    # perform multiplication
    ww = (z1 + x1) * (x2 + y2)
    yy = (w1 - y1) * (w2 + z2)
    zz = (w1 + y1) * (w2 - z2)
    xx = ww + yy + zz
    qq = 0.5 * (xx + (z1 - x1) * (x2 - y2))
    w = qq - ww + (z1 - y1) * (y2 - z2)
    x = qq - xx + (x1 + w1) * (x2 + w2)
    y = qq - yy + (w1 - x1) * (y2 + z2)
    z = qq - zz + (z1 + y1) * (w2 - x2)

    return np.stack([w, x, y, z], axis=-1).reshape(shape)



from typing import Callable

import numpy as np
import bpy
import bpy_types
from mathutils import Vector

from .motion_sequence import MotionSequence


C = bpy.context
D = bpy.data
O = bpy.ops


"""
Generic quick references:

# getting an armature
armature = D.objects.get("Armature")

# get pose bones
bones = armature.pose.bones

# get bone location
bones.get("bone_name").location

# get bone rotation
bones.get("bone_name").rotation_quaternion

# get bone rotation matrix
bones.get("bone_name").matrix

# get bone scale
bones.get("bone_name").scale

# get edit bones
armature.data.bones
"""


def cleanup_usd_axis_display(display_size: float = 0.01) -> None:
    """
    This function shrinks the axis display of the imported IsaacLab USD object
    to make this looks cleaner.

    Args:
        display_size: the size of the axes
    """
    for object in D.objects:
        if "visuals" in object.name or "collisions" in object.name:
            object.empty_display_size = display_size


def set_scene_fps(fps: int) -> None:
    """
    Set the scene FPS.
    """
    C.scene.render.fps = fps
    C.scene.render.fps_base = 1.0


def set_scene_animation_range(start: int, end: int) -> None:
    """
    Set the animation range for the current scene.

    Args:
        start: The start frame.
        end: The end frame.
    """
    C.scene.frame_start = start
    C.scene.frame_end = end


def set_armature_to_rest(armature: bpy_types.Object) -> None:
    """
    Set the armature to rest pose.
    """
    armature.data.pose_position = "REST"


def set_armature_to_pose(armature: bpy_types.Object) -> None:
    """
    Set the armature to animation pose.
    """
    armature.data.pose_position = "POSE"


def set_bones_to_1d_rotation(armature: bpy_types.Object) -> None:
    """
    Set the bones to 1D rotation mode, and allow only Y-axis rotation (along the bone axis).
    Use this function to convert arbitrary armature to "realistic" robot armature with revolute joints.

    Args:
        armature: The armature object.
    """

    for bone in armature.pose.bones:
        print(f"found bone {bone.name}")
        # set to XYZ rotation mode
        bone.rotation_mode = "XYZ"

        # allow only Y-axis rotation
        bone.lock_rotation[0] = False
        bone.lock_rotation[1] = False
        bone.lock_rotation[2] = False

        # allow only Y-axis rotation in IK
        bone.lock_ik_x = False
        bone.lock_ik_y = False
        bone.lock_ik_z = False


def build_body_motion_data(
    armature: bpy_types.Object,
    mapping: dict[str, tuple[str, Callable]],
    scaling_ratio: float = 1.0,
) -> MotionSequence:
    """
    Build rigid body motion data from the source armature.
    The dof motion data is not included in this function, which will be initialized 
    as a properly-dimensioned zero array.

    Args:
        armature: The source armature object.
        mapping: The mapping from source armature bone names to target rigid body names.
        scaling_ratio: The scaling ratio of the armature.

    Returns:
        A MotionSequence object containing the motion data.
    """
    fps_float = C.scene.render.fps / C.scene.render.fps_base
    start_frame = C.scene.frame_start
    end_frame = C.scene.frame_end

    assert end_frame >= start_frame, f"Frame range is invalid: {start_frame} to {end_frame}"

    link_names = list(mapping.keys())

    n_frames = end_frame - start_frame + 1

    motion = MotionSequence(
        num_frames=n_frames,
        dof_names=[],
        body_names=link_names,
        fps=fps_float,
    )

    # used to calculate angular velocities
    body_rotations_euler = np.zeros((n_frames, len(link_names), 3), dtype=np.float32)

    # === extract motion data ===
    for frame in range(n_frames):
        # navigate to the corresponding frame
        bpy.context.scene.frame_set(start_frame + frame)
        bpy.context.view_layer.update()

        # force UI update to update bone pose matrix
        bpy.ops.wm.redraw_timer(type="DRAW_WIN_SWAP", iterations=1)

        # read bone positions
        for idx, name in enumerate(link_names):
            entry = mapping.get(name)
            if not entry:
                print(f"WARNING: cannot find link mapping for {name}")
                continue

            source_bone_name, mapping_function = entry

            source_bone = armature.pose.bones.get(source_bone_name)
            if not source_bone:
                print(f"WARNING: cannot find source bone {source_bone_name}")
                continue

            motion._body_positions[frame, idx, :] = mapping_function(source_bone)

            matrix = source_bone.matrix
            motion._body_rotations[frame, idx, :] = matrix.to_quaternion()
            body_rotations_euler[frame, idx, :] = matrix.to_euler()

        print(f"Processing: #{frame}/{n_frames} ({frame / n_frames * 100:.2f}%)", end="\r")

    # === post-process motion data ===
    # cancel first frame global offset
    offset_x = np.mean(motion._body_positions[0, :, 0])
    offset_y = np.mean(motion._body_positions[0, :, 1])
    motion._body_positions[:, :, 0] -= offset_x
    motion._body_positions[:, :, 1] -= offset_y

    # in Blender, scaling the armature does not scale the retreived bone position, so we need to
    # manually apply the scaling to the sampled data here.
    motion._body_positions[:] *= scaling_ratio

    # calculate velocities
    motion._body_linear_velocities[1:] = np.diff(motion._body_positions, axis=0) / (1. / fps_float)

    # calculate angular velocities
    # handle euler angle discontinuity
    # TODO: this is not quite correct, we need to use quaternions to calculate angular velocities
    body_rotations_euler = np.unwrap(body_rotations_euler, axis=0)
    motion._body_angular_velocities[1:] = np.diff(body_rotations_euler, axis=0) / (1. / fps_float)

    # handle euler angle wrapping
    motion._body_angular_velocities = np.unwrap(motion._body_angular_velocities, axis=0)

    print(f"Done generating {n_frames} frames ({n_frames / fps_float:.2f} seconds)")

    return motion


def construct_skeleton_tree():
    skeleton_tree_usd = {
        "node_names": [  # joint names
            "pelvis",                     # 0
            "waist_yaw",                  # 1
            "waist_roll",                 # 2
            "waist_pitch",                # 3
            "neck_yaw",                   # 4
            "neck_roll",                  # 5
            "neck_pitch",                 # 6
            "arm_left_shoulder_pitch",    # 7
            "arm_left_shoulder_roll",     # 8
            "arm_left_shoulder_yaw",      # 9
            "arm_left_elbow_pitch",       # 10
            "arm_left_wrist_yaw",         # 11
            "arm_left_wrist_roll",        # 13
            "arm_left_wrist_pitch",       # 12
            "arm_right_shoulder_pitch",   # 14
            "arm_right_shoulder_roll",    # 15
            "arm_right_shoulder_yaw",     # 16
            "arm_right_elbow_pitch",      # 17
            "arm_right_wrist_yaw",        # 18
            "arm_right_wrist_roll",       # 20
            "arm_right_wrist_pitch",      # 19
            "leg_left_hip_pitch",         # 21
            "leg_left_hip_roll",          # 22
            "leg_left_hip_yaw",           # 23
            "leg_left_knee_pitch",        # 24
            "leg_left_ankle_yaw",         # 25
            "leg_left_ankle_pitch",       # 26
            "leg_left_ankle_roll",        # 27
            "leg_right_hip_pitch",        # 28
            "leg_right_hip_roll",         # 29
            "leg_right_hip_yaw",          # 30
            "leg_right_knee_pitch",       # 31
            "leg_right_ankle_yaw",        # 32
            "leg_right_ankle_pitch",      # 33
            "leg_right_ankle_roll",       # 34
        ],
        "link_names": [
            "pelvis",                     # 0
            "chest_yaw",                  # 1
            "chest_roll",                 # 2
            "chest_pitch",                # 3
            "head_yaw",                   # 4
            "head_roll",                  # 5
            "head_pitch",                 # 6
            "arm_left_upper_pitch",       # 7
            "arm_left_upper_roll",        # 8
            "arm_left_upper_yaw",         # 9
            "arm_left_forearm_pitch",     # 10
            "arm_left_hand_yaw",          # 11
            "arm_left_hand_roll",         # 12
            "arm_left_hand_pitch",        # 13
            "arm_right_upper_pitch",      # 14
            "arm_right_upper_roll",       # 15
            "arm_right_upper_yaw",        # 16
            "arm_right_forearm_pitch",    # 17
            "arm_right_hand_yaw",         # 18
            "arm_right_hand_roll",        # 19
            "arm_right_hand_pitch",       # 20
            "leg_left_thigh_pitch",       # 21
            "leg_left_thigh_roll",        # 22
            "leg_left_thigh_yaw",         # 23
            "leg_left_calf_pitch",        # 24
            "leg_left_foot_yaw",          # 25
            "leg_left_foot_pitch",        # 26
            "leg_left_foot_roll",         # 27
            "leg_right_thigh_pitch",      # 28
            "leg_right_thigh_roll",       # 29
            "leg_right_thigh_yaw",        # 30
            "leg_right_calf_pitch",       # 31
            "leg_right_foot_yaw",         # 32
            "leg_right_foot_pitch",       # 33
            "leg_right_foot_roll",        # 34
        ],
        "parent_indices": {
            "arr": np.array([
                -1,
                0, 1, 2, 3, 4, 5,
                3, 7, 8, 9, 10, 11, 12,
                3, 14, 15, 16, 17, 18, 19,
                0, 21, 22, 23, 24, 25, 26,
                0, 28, 29, 30, 31, 32, 33
            ]),
            "context": {"dtype": "int64"}
            },
        "global_translations": [],
        "bone_orientations": [
            [ .0,  .1,  .0],
            [ .0,  .0,  .1],
            [ .1,  .0,  .0],
            [ .0,  .1,  .0],
            [ .0,  .0,  .1],
            [ .1,  .0,  .0],
            [ .0,  .1,  .0],

            [ .0,  .1,  .0],
            [ .1,  .0,  .0],
            [ .0, -.1*np.cos(np.deg2rad(60)),  .1*np.sin(np.deg2rad(60))],
            [ .0,  .1*np.sin(np.deg2rad(60)),  .1*np.cos(np.deg2rad(60))],
            [ .0, -.1*np.cos(np.deg2rad(60)),  .1*np.sin(np.deg2rad(60))],
            [ .0,  .1*np.sin(np.deg2rad(60)),  .1*np.cos(np.deg2rad(60))],
            [ .1,  .0,  .0],
            [ .0,  .1,  .0],
            [ .1,  .0,  .0],
            [ .0,  .1*np.cos(np.deg2rad(60)),  .1*np.sin(np.deg2rad(60))],
            [ .0,  .1*np.sin(np.deg2rad(60)), -.1*np.cos(np.deg2rad(60))],
            [ .0,  .1*np.cos(np.deg2rad(60)),  .1*np.sin(np.deg2rad(60))],
            [ .0,  .1*np.sin(np.deg2rad(60)), -.1*np.cos(np.deg2rad(60))],
            [ .1,  .0,  .0],

            [ .0,  .1,  .0],
            [ .1,  .0,  .0],
            [ .0,  .0,  .1],
            [ .0,  .1,  .0],
            [ .0,  .0,  .1],
            [ .0,  .1,  .0],
            [ .1,  .0,  .0],

            [ .0,  .1,  .0],
            [ .1,  .0,  .0],
            [ .0,  .0,  .1],
            [ .0,  .1,  .0],
            [ .0,  .0,  .1],
            [ .0,  .1,  .0],
            [ .1,  .0,  .0],
        ],
    }

    for link_name in skeleton_tree_usd["link_names"]:
        skeleton_tree_usd["global_translations"].append(np.array(D.objects.get(link_name).location))

    # skeleton_tree_usd["global_translations"] += np.array([0, 0, 0.6])

    return skeleton_tree_usd


class SkeletonTree:
    skeleton_tree_a_pose = {
        "node_names": [
            "pelvis",               # 0
            "chest",                # 1
            "head",                 # 2
            "arm_left_upper",       # 3
            "arm_left_lower",       # 4
            "arm_left_hand",        # 5
            "arm_right_upper",      # 6
            "arm_right_lower",      # 7
            "arm_right_hand",       # 8
            "leg_left_upper",       # 9
            "leg_left_lower",       # 10
            "leg_left_foot",        # 11
            "leg_right_upper",      # 12
            "leg_right_lower",      # 13
            "leg_right_foot",       # 14
        ],
        "parent_indices": {
            "arr": np.array([-1, 0, 1, 1, 3, 4, 1, 6, 7, 0, 9, 10, 0, 12, 13]),
            "context": {"dtype": "int64"}
            },
        "local_translations": [
            [  .000,   .000,  0.600],
            [  .000,   .000,  0.100],
            [  .000,   .000,  0.300],
            [-0.018,  0.066,  0.218],
            [  .000,  0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000,  0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [-0.018, -0.066,  0.218],
            [  .000, -0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000, -0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000,  0.084,   .000],
            [  .000, -0.030, -0.255],
            [-0.023,   .000, -0.293],
            [  .000, -0.084,   .000],
            [  .000,  0.030, -0.255],
            [-0.023,   .000, -0.293],
        ],
        "bone_orientations": [
            [  .000,   .000,  0.100],
            [  .000,   .000,  0.218],
            [  .000,   .000,  0.100],
            [  .000,  0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000,  0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000,  0.100*np.cos(np.deg2rad(60)), -0.100*np.sin(np.deg2rad(60))],
            [  .000, -0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000, -0.150*np.cos(np.deg2rad(60)), -0.150*np.sin(np.deg2rad(60))],
            [  .000, -0.100*np.cos(np.deg2rad(60)), -0.100*np.sin(np.deg2rad(60))],
            [  .000,   .000, -0.250],
            [-0.023,   .000, -0.290],
            [ 0.066,   .000, -0.056],
            [  .000,   .000, -0.250],
            [-0.023,   .000, -0.290],
            [ 0.066,   .000, -0.056],
        ],
    }

    skeleton_tree_t_pose = {
        "node_names": [
            "pelvis",               # 0
            "chest",                # 1
            "head",                 # 2
            "arm_left_upper",       # 3
            "arm_left_lower",       # 4
            "arm_left_hand",        # 5
            "arm_right_upper",      # 6
            "arm_right_lower",      # 7
            "arm_right_hand",       # 8
            "leg_left_upper",       # 9
            "leg_left_lower",       # 10
            "leg_left_foot",        # 11
            "leg_right_upper",      # 12
            "leg_right_lower",      # 13
            "leg_right_foot",       # 14
        ],
        "parent_indices": {
            "arr": np.array([-1, 0, 1, 1, 3, 4, 1, 6, 7, 0, 9, 10, 0, 12, 13]),
            "context": {"dtype": "int64"}
            },
        "local_translations": [
            [  .000,   .000,  0.600],
            [  .000,   .000,  0.100],
            [  .000,   .000,  0.300],
            [-0.018,  0.066,  0.218],
            [  .000,  0.150,   .000],
            [  .000,  0.150,   .000],
            [-0.018, -0.066,  0.218],
            [  .000, -0.150,   .000],
            [  .000, -0.150,   .000],
            [  .000,  0.084,   .000],
            [  .000, -0.030, -0.255],
            [-0.023,   .000, -0.293],
            [  .000, -0.084,   .000],
            [  .000,  0.030, -0.255],
            [-0.023,   .000, -0.293],
        ],
        "bone_orientations": [
            [  .000,   .000,  0.100],
            [  .000,   .000,  0.218],
            [  .000,   .000,  0.100],
            [  .000,  0.150,   .000],
            [  .000,  0.150,   .000],
            [  .000,  0.100,   .000],
            [  .000, -0.150,   .000],
            [  .000, -0.150,   .000],
            [  .000, -0.100,   .000],
            [  .000,   .000, -0.250],
            [-0.023,   .000, -0.290],
            [ 0.066,   .000, -0.056],
            [  .000,   .000, -0.250],
            [-0.023,   .000, -0.290],
            [ 0.066,   .000, -0.056],
        ],
    }


def build_armature(skeleton_tree: dict, armature_name="Armature"):
    armature = D.objects.get(armature_name)

    # delete the existing armature, if any
    if armature:
        O.object.mode_set(mode="OBJECT")
        armature.select_set(True)
        O.object.delete()


    O.object.armature_add(enter_editmode=False, align="WORLD", scale=(1, 1, 1))
    armature = C.active_object
    armature.name = armature_name

    # switch to edit mode
    O.object.mode_set(mode="EDIT")
    edit_bones = armature.data.edit_bones

    # reconfigure root bone
    root = edit_bones[0]
    root.name = "root"
    root.tail = Vector([0.5, 0, 0])


    for name in skeleton_tree["node_names"]:
        edit_bones.new(name)

    for i, name in enumerate(skeleton_tree["node_names"]):
        parent_idx = skeleton_tree["parent_indices"]["arr"][i]

        bone = edit_bones.get(name)

        if skeleton_tree.get("local_translations"):
            # use local translation
            if parent_idx == -1:
                bone.head = Vector(skeleton_tree["local_translations"][i])
                bone.parent = root
            else:
                parent_bone_name = skeleton_tree["node_names"][parent_idx]
                parent_bone = edit_bones.get(parent_bone_name)

                bone.head = parent_bone.head + Vector(skeleton_tree["local_translations"][i])
                bone.parent = parent_bone

        else:
            # use global translation
            if parent_idx == -1:
                bone.parent = root
            else:
                parent_bone_name = skeleton_tree["node_names"][parent_idx]
                parent_bone = edit_bones.get(parent_bone_name)

                bone.parent = parent_bone

            bone.head = Vector(skeleton_tree["global_translations"][i])

        bone.tail = bone.head + Vector(skeleton_tree["bone_orientations"][i])

    O.object.mode_set(mode="OBJECT")

    bones = D.objects.get("Armature").pose.bones
    for bone in bones:
        bone.rotation_mode = "XYZ"

def bind_to_armature(skeleton_tree: dict):

    armature = D.objects.get("Armature")

    for idx, link_name in enumerate(skeleton_tree["link_names"]):
        frame = D.objects.get(link_name)
        bone_name = skeleton_tree["node_names"][idx]

        print(f"binding {link_name} to {bone_name}")

        # the use of matrix world is to maintain the original transform
        # i.e. the equivalent of "Keep Transform" GUI option

        # save original world matrix
        matrix_world = frame.matrix_world.copy()

        frame.parent = armature
        frame.parent_bone = bone_name
        frame.parent_type = "BONE"

        # restore world matrix to maintain global transform
        frame.matrix_world = matrix_world.copy()


def load_replay(skeleton_tree: dict, armature: bpy_types.Object, data_path: str):
    # TODO: refactor the replay motion format to use the same format as the motion data
    data = np.load(data_path)

    fps = data["fps"]
    joint_order = data["joint_order"].tolist()
    root_positions = data["root_positions"]
    root_quaternions = data["root_quaternions"]
    joint_positions = data["joint_positions"]


    print(joint_positions.shape)

    n_frames = joint_positions.shape[0]
    n_dof = joint_positions.shape[-1]

    bpy.context.scene.frame_start = 0
    bpy.context.scene.frame_end = n_frames

    for frame in range(n_frames):
        bpy.context.scene.frame_set(frame)
        bpy.context.view_layer.update()

        root = armature.pose.bones.get("pelvis")
        root.location = root_positions[frame]
        # root.location[2] -= 0.6
        root.keyframe_insert(data_path="location", frame=frame)

        root.rotation_mode = "QUATERNION"
        root.rotation_quaternion = root_quaternions[frame]
        root.keyframe_insert(data_path="rotation_quaternion", frame=frame)

        for joint_idx, joint_name in enumerate(joint_order):
            bone_name = joint_name.replace("_joint", "")
            bone = armature.pose.bones.get(bone_name)
            # ensure using Euler angles
            bone.rotation_mode = "XYZ"
            bone.rotation_euler[1] = joint_positions[frame, joint_idx]

            # insert rotation_euler keyframe, for index 1 (Y-axis)
            bone.keyframe_insert(data_path="rotation_euler", index=1, frame=frame)

        print(f"Processing: #{frame}/{n_frames} ({frame / n_frames * 100:.2f}%)")




# only thing need to be cautious about is that the first entry must be the root.


class UnitreeG1Mapping:
    actorcore = {
        "pelvis"                    : ("CC_Base_Pelvis",        lambda b: b.head),  # 0
        "left_shoulder_roll_link"   : ("CC_Base_L_Clavicle",    lambda b: b.tail),  # 1
        "left_elbow_link"           : ("CC_Base_L_Forearm",     lambda b: b.head),  # 2
        "left_wrist_pitch_link"     : ("CC_Base_L_Hand",        lambda b: b.head),  # 3
        "left_rubber_hand"          : ("CC_Base_L_Hand",        lambda b: b.tail),  # 4
        "right_shoulder_roll_link"  : ("CC_Base_R_Clavicle",    lambda b: b.tail),  # 5
        "right_elbow_link"          : ("CC_Base_R_Forearm",     lambda b: b.head),  # 6
        "right_wrist_pitch_link"    : ("CC_Base_R_Hand",        lambda b: b.head),  # 7
        "right_rubber_hand"         : ("CC_Base_R_Hand",        lambda b: b.tail),  # 8
        "left_hip_roll_link"        : ("CC_Base_L_Thigh",       lambda b: b.head),  # 9
        "left_knee_link"            : ("CC_Base_L_Calf",        lambda b: b.head),  # 10
        "left_ankle_roll_link"      : ("CC_Base_L_Foot",        lambda b: b.head),  # 11
        "right_hip_roll_link"       : ("CC_Base_R_Thigh",       lambda b: b.head),  # 12
        "right_knee_link"           : ("CC_Base_R_Calf",        lambda b: b.head),  # 13
        "right_ankle_roll_link"     : ("CC_Base_R_Foot",        lambda b: b.head),  # 14
        "torso_link"                : ("CC_Base_Spine02",       lambda b: b.head),  # 15
        "head_link"                 : ("CC_Base_Head",          lambda b: b.tail),  # 16
    }
    meshcapade = {
        "pelvis"                    : ("pelvis",        lambda b: b.head),
        "left_shoulder_roll_link"   : ("upperarm_l",    lambda b: b.head),
        "left_elbow_link"           : ("lowerarm_l",    lambda b: b.head),
        "left_wrist_pitch_link"     : ("hand_l",        lambda b: b.head),
        "left_rubber_hand"          : ("hand_l",        lambda b: b.tail),
        "right_shoulder_roll_link"  : ("upperarm_r",    lambda b: b.head),
        "right_elbow_link"          : ("lowerarm_r",    lambda b: b.head),
        "right_wrist_pitch_link"    : ("hand_r",        lambda b: b.head),
        "right_rubber_hand"         : ("hand_r",        lambda b: b.tail),
        "torso_link"                : ("torso",         lambda b: b.head),
        "head_link"                 : ("head",          lambda b: b.head),
        "left_hip_roll_link"        : ("thigh_l",       lambda b: b.head),
        "left_knee_link"            : ("calf_l",        lambda b: b.head),
        "left_ankle_pitch_link"     : ("foot_l",        lambda b: b.head),
        "left_ankle_roll_link"      : ("ball_l",        lambda b: b.head),
        "right_hip_roll_link"       : ("thigh_r",       lambda b: b.head),
        "right_knee_link"           : ("calf_r",        lambda b: b.head),
        "right_ankle_pitch_link"    : ("foot_r",        lambda b: b.head),
        "right_ankle_roll_link"     : ("ball_r",        lambda b: b.head),
    }
    mmd_yyb = {
        "pelvis"                    : ("下半身",     lambda b: b.tail),  # 0
        "left_shoulder_roll_link"   : ("腕.L",      lambda b: b.head),  # 1
        "left_elbow_link"           : ("ひじ.L",    lambda b: b.head),  # 2
        "left_rubber_hand"          : ("手首.L",    lambda b: b.head),  # 3
        "right_shoulder_roll_link"  : ("腕.R",      lambda b: b.head),  # 4
        "right_elbow_link"          : ("ひじ.R",    lambda b: b.head),  # 5
        "right_rubber_hand"         : ("手首.R",    lambda b: b.head),  # 6
        "torso_link"                : ("上半身",    lambda b: b.head),  # 7
        "left_hip_roll_link"        : ("足.L",      lambda b: b.head),  # 8
        "left_knee_link"            : ("ひざ.L",    lambda b: b.head),  # 9
        "left_ankle_roll_link"      : ("足首.L",    lambda b: b.head),  # 10
        "right_hip_roll_link"       : ("足.R",      lambda b: b.head),  # 11
        "right_knee_link"           : ("ひざ.R",    lambda b: b.head),  # 12
        "right_ankle_roll_link"     : ("足首.R",    lambda b: b.head),  # 13
        "head_link"                 : ("頭",        lambda b: b.head),  # 14
    }


class AirDraftMapping:
    mixamo = {
        "pelvis"            : ("mixamorig:Hips",            lambda b: b.head),
        "left_shoulder"     : ("mixamorig:LeftArm",         lambda b: b.head),
        "left_elbow"        : ("mixamorig:LeftForeArm",     lambda b: b.head),
        "left_hand"         : ("mixamorig:LeftHand",        lambda b: b.head),
        "right_shoulder"    : ("mixamorig:RightArm",        lambda b: b.head),
        "right_elbow"       : ("mixamorig:RightForeArm",    lambda b: b.head),
        "right_hand"        : ("mixamorig:RightHand",       lambda b: b.head),
        "left_hip"          : ("mixamorig:LeftUpLeg",       lambda b: b.head),
        "left_knee"         : ("mixamorig:LeftLeg",         lambda b: b.head),
        "left_foot"         : ("mixamorig:LeftToeBase",     lambda b: b.head),
        "right_hip"         : ("mixamorig:RightUpLeg",      lambda b: b.head),
        "right_knee"        : ("mixamorig:RightLeg",        lambda b: b.head),
        "right_foot"        : ("mixamorig:RightToeBase",    lambda b: b.head),
        "chest"             : ("mixamorig:Spine1",          lambda b: b.head),
        "head"              : ("mixamorig:Head",            lambda b: b.head),
    }
    mmd_yyb = {
        "pelvis"            : ("腰",        lambda b: b.head),
        "left_shoulder"     : ("腕.L",      lambda b: b.head),
        "left_elbow"        : ("腕.L",      lambda b: b.tail),
        "left_hand"         : ("ひじ.L",    lambda b: b.tail),
        "right_shoulder"    : ("腕.R",      lambda b: b.head),
        "right_elbow"       : ("腕.R",      lambda b: b.tail),
        "right_hand"        : ("ひじ.R",    lambda b: b.tail),
        "left_hip"          : ("足.L",      lambda b: b.head),
        "left_knee"         : ("足.L",      lambda b: b.tail),
        "left_foot"         : ("ひざ.L",    lambda b: b.tail),
        "right_hip"         : ("足.R",      lambda b: b.head),
        "right_knee"        : ("足.R",      lambda b: b.tail),
        "right_foot"        : ("ひざ.R",    lambda b: b.tail),
        "chest"             : ("上半身",    lambda b: b.head),
        "head"              : ("頭",        lambda b: b.head),
    }
